# coding=utf-8
from __future__ import absolute_import, print_function

import os

from suanpan import api, path
from suanpan.arguments import String
from suanpan.tools import ToolComponent as tc
from suanpan.utils import json


@tc.param(String(key="app", required=True))
@tc.param(String(key="dist", default=os.path.join("tmp", "graph")))
def SPAppGraphDownload(context):
    args = context.args
    filepath = args.dist if args.dist.endswith(".json") else os.path.join(args.dist, f"{args.app}.json")
    path.mkdirs(filepath, parent=True)
    json.dump(api.app.getAppGraph(args.app), filepath)


if __name__ == "__main__":
    SPAppGraphDownload()  # pylint: disable=no-value-for-parameter



# coding=utf-8
from __future__ import absolute_import, print_function

import os

from suanpan import error, utils
from suanpan.log import logger
from suanpan.objects import HasName
from suanpan.utils import env, json

DEFAULT_MAX_VALUE_LENGTH = 120


class Arg(HasName):
    MAX_VALUE_LENGTH = DEFAULT_MAX_VALUE_LENGTH

    def __init__(self, key, **kwargs):
        self.alias = kwargs.pop("alias", None)
        self.key = key
        self.required = kwargs.pop("required", False)
        self.default = kwargs.pop("default", None)
        self.type = kwargs.pop("type", str)
        self.value = None

        self.kwargs = self.cleanParams(kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def isSet(self):
        return self.required or self.default != self.value

    def load(self, args):
        self.value = args.get(self.key)
        if self.value is None:
            self.value = env.get(self.envKeyFormat(self.key))
        if self.required and self.value is None and self.default is None:
            raise error.ArgumentRequiredError(f"{self.key} is required")
        return self

    def format(self):
        if self.value is None:
            self.value = self.default
        else:
            try:
                self.value = self.transform(self.value)
            except Exception:
                raise error.ArgumentTypeError(
                    f"({self.name}) {self.key}: {utils.shorten(self.value, maxlen=self.MAX_VALUE_LENGTH)}"
                )
        self.logLoaded(self.value)
        return self

    def transform(self, value):
        return self.type(value)

    def clean(self):  # pylint: disable=unused-argument
        return self

    def save(self, result):  # pylint: disable=unused-argument
        self.logSaved(result.value)
        return result.value

    def cleanParams(self, params):
        return {k: v for k, v in params.items() if not k.startswith("_")}

    @property
    def keyString(self):
        return self.alias or self.key

    def logLoaded(self, value):
        logger.info(
            f"({self.name}) {self.keyString} loaded: {utils.shorten(value, maxlen=self.MAX_VALUE_LENGTH)}"
        )

    def logSaved(self, value):
        logger.info(
            f"({self.name}) {self.keyString} saved: {utils.shorten(value, maxlen=self.MAX_VALUE_LENGTH)}"
        )

    def fixGlobalKey(self, key):
        return key.replace("-", "_")

    def envKeyFormat(self, key):
        return f"SP_{self.fixGlobalKey(key).upper()}"

    def getOutputTmpValue(self, *args):
        pass

    def getOutputTmpArg(self, *args):
        value = self.getOutputTmpValue(  # pylint: disable=assignment-from-no-return
            *args
        )
        return (f"--{self.key}", value) if value is not None else tuple()


class String(Arg):
    pass

class Int(Arg):
    @classmethod
    def transform(cls, value):
        return int(value)

class Float(Arg):
    @classmethod
    def transform(cls, value):
        return float(value)


class Bool(Arg):
    def __init__(self, key, **kwargs):
        kwargs.setdefault("default", False)
        super(Bool, self).__init__(key, **kwargs)

    @classmethod
    def transform(cls, value):
        if value.lower() in ("yes", "true", "t", "y"):
            return True
        if value.lower() in ("no", "false", "f", "n"):
            return False
        raise error.ArgumentError(cls.__name__)


class List(Arg):
    def transform(self, value):
        return [i.strip() for i in value.split(",") if i.strip()]


class ListOfString(List):
    pass


class ListOfInt(List):
    def transform(self, value):
        items = super(ListOfInt, self).transform(value)
        return [Int.transform(i) for i in items]


class ListOfFloat(List):
    def transform(self, value):
        items = super(ListOfFloat, self).transform(value)
        return [Float.transform(i) for i in items]


class ListOfBool(List):
    def transform(self, value):
        items = super(ListOfBool, self).transform(value)
        return [Bool.transform(i) for i in items]


class Json(String):
    def transform(self, value):
        if value is not None:
            value = json.loads(value)
        return value

    def save(self, result):
        super(Json, self).save(result)
        return json.dumps(result.value)


class IntOrFloat(Arg):
    @classmethod
    def transform(cls, value):
        return Float.transform(value) if "." in value else Int.transform(value)


class IntFloatOrString(Arg):
    def transform(self, value):
        try:
            Float.transform(value)
            return IntOrFloat.transform(value)
        except ValueError:
            return value
        except TypeError:
            return value


class BoolOrString(Arg):
    def transform(self, value):
        return value if value == "auto" else Bool.transform(value)


class BoolOrInt(Arg):
    def transform(self, value):
        try:
            return Bool.transform(value)
        except error.ArgumentError:
            return Int.transform(value)


class StringOrListOfFloat(ListOfFloat):
    def transform(self, value):
        if "," in value:
            return super(StringOrListOfFloat, self).transform(value)
        return value



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.docker import DockerComponent as dc
from suanpan.docker.arguments import Json
from suanpan.utils import json


@dc.input(Json(key="inputData1", required=True))
@dc.output(Json(key="outputData1", required=True))
@dc.param(Json(key="param1", required=True))
def Demo(context):
    args = context.args
    return args.inputData1


if __name__ == "__main__":
    Demo(  # pylint: disable=all
        inputData1="input_test", outputData1="output_test", param1=json.dumps({"b": 2})
    )



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.g import g
from suanpan.run import cli, env, run

__version__ = "0.12.14"


if __name__ == "__main__":
    print(f"Suanpan SDK (ver: {__version__})")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import String
from suanpan.storage import storage
from suanpan.tools import ToolComponent as tc


@tc.param(String(key="remote", required=True))
def SPStorageRemove(context):
    args = context.args
    storage.remove(args.remote)


if __name__ == "__main__":
    SPStorageRemove()  # pylint: disable=no-value-for-parameter



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class ComponentError(base.Error):
    MESSAGE = "Suanpan Component Error: {}"


class ComponentHandlerNotSet(ComponentError):
    MESSAGE = "Suanpan Component Handler Not Set: {}"


class ComponentHandlerNotCallable(ComponentError):
    MESSAGE = "Suanpan Component Handler Not Callable: {}"



import suanpan
from suanpan.app import app
from suanpan.app.arguments import Folder, String, Int

@app.input(Folder(key="inputData1"))
@app.output(Folder(key="outputData1"))
@app.param(String(key="param1"))
def test1(context):
    args = context.args
    print(args.param1)
    return args.outputData1


@app.trigger.interval(1)
@app.trigger.output(Folder(key="outputData1"))
@app.trigger.param(String(key="param2"))
def test2(context):
    args = context.args
    print(args.param2)
    app.send({"qwe": 1}, args=[Int(key="qwe")])
    return args.outputData1


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, print_function


class Error(Exception):
    MESSAGE = "Suanpan Error: {}"

    def __init__(self, *args, **kwargs):
        super(Error, self).__init__(self.MESSAGE.format(*args, **kwargs))



# coding=utf-8
from __future__ import absolute_import, print_function

import argparse
import contextlib
import copy
import sys

from suanpan.imports import imports
from suanpan.utils import env as spenv


def run(component, *args, **kwargs):
    if isinstance(component, str):
        component = f"{component[:-3]}.app" if component.endswith(".py") else component
        component = imports(component)
    with env(**kwargs.pop("env", {})):
        return component.start(*args, **kwargs)


@contextlib.contextmanager
def env(**kwargs):
    old = copy.deepcopy(spenv.environ)
    spenv.update(kwargs)
    yield spenv.environ
    spenv.update(old)


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("component")
    _args, _rest = parser.parse_known_args()

    sys.argv = sys.argv[:1]
    return run(_args.component, *_rest)


if __name__ == "__main__":
    cli()



# coding=utf-8
from __future__ import absolute_import, print_function

import functools
import itertools

from suanpan import error
from suanpan.dw import dw
from suanpan.interfaces import HasArguments
from suanpan.log import logger
from suanpan.mq import mq
from suanpan.mstorage import mstorage
from suanpan.storage import storage


class HasBaseServices(HasArguments):
    ENABLED_BASE_SERVICES = {"mq", "mstorage", "dw", "storage"}

    def setBackendIfHasArgs(*keys):  # pylint: disable=no-method-argument
        def _wrap(func):
            @functools.wraps(func)
            def _dec(cls, sargs, *args, **kwargs):
                return (
                    func(cls, sargs, *args, **kwargs)
                    if all(getattr(sargs, key, None) is not None for key in keys)
                    else None
                )

            return _dec

        return _wrap

    @classmethod
    def setBaseServices(cls, args):
        mapping = {
            "mq": cls.setMQ,
            "mstorage": cls.setMStorage,
            "dw": cls.setDataWarehouse,
            "storage": cls.setStorage,
        }
        setters = {
            service: mapping.get(service) for service in cls.ENABLED_BASE_SERVICES
        }
        if not all(setters.values()):
            raise error.UnknownBaseServicesError(
                set(cls.ENABLED_BASE_SERVICES) - set(mapping.keys())
            )
        backends = ((service, setter(args)) for service, setter in setters.items())
        baseServices = {service: backend for service, backend in backends if backend}
        baseServicesString = ", ".join(baseServices.keys())
        logger.debug(f"Base services init: {baseServicesString}")
        return baseServices

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super(HasBaseServices, self).getGlobalArguments(*args, **kwargs)
        return arguments + self.getBaseServicesArguments()

    @classmethod
    def getBaseServicesArguments(cls):
        mapping = {
            "mq": mq.ARGUMENTS,
            "mstorage": mstorage.ARGUMENTS,
            "dw": dw.ARGUMENTS,
            "storage": storage.ARGUMENTS,
        }
        arguments = {
            service: mapping.get(service) for service in cls.ENABLED_BASE_SERVICES
        }
        if not all(arguments.values()):
            raise error.UnknownBaseServicesError(
                set(cls.ENABLED_BASE_SERVICES) - set(mapping.keys())
            )
        return list(itertools.chain(*arguments.values()))

    @classmethod
    @setBackendIfHasArgs("mq_type")
    def setMQ(cls, argsDict):
        return mq.setBackend(**cls.defaultArgumentsFormat(argsDict, mq.ARGUMENTS))

    @classmethod
    @setBackendIfHasArgs("mstorage_type")
    def setMStorage(cls, argsDict):
        return mstorage.setBackend(
            **cls.defaultArgumentsFormat(argsDict, mstorage.ARGUMENTS)
        )

    @classmethod
    @setBackendIfHasArgs("dw_type")
    def setDataWarehouse(cls, argsDict):
        return dw.setBackend(**cls.defaultArgumentsFormat(argsDict, dw.ARGUMENTS))

    @classmethod
    @setBackendIfHasArgs("storage_type")
    def setStorage(cls, argsDict):
        return storage.setBackend(**cls.defaultArgumentsFormat(argsDict, storage.ARGUMENTS))



# coding=utf-8
from __future__ import absolute_import, print_function

from tabulate import tabulate

from suanpan.utils.term import graph


def chart(data, labels=None, colors=None, **kwargs):
    args = {
        "filename": "-",
        "title": None,
        "width": 50,
        "format": "{:<5.2f}",
        "suffix": "",
        "no_labels": not bool(labels),
        "color": None,
        "vertical": False,
        "stacked": False,
        "different_scale": False,
        "calendar": False,
        "start_dt": None,
        "custom_tick": "",
        "delim": "",
        "verbose": False,
        "version": False,
    }
    args.update(kwargs)
    colors = colors or []
    labels = labels or [None] * len(data)
    if args["calendar"]:
        graph.calendar_heatmap(data, labels, args)
    else:
        graph.chart(colors, data, args, labels)


def table(data, headers=(), theme="plain"):
    print(tabulate(data, headers=headers, tablefmt=theme))



# coding=utf-8
from __future__ import absolute_import, division, print_function

import os
import time
import webbrowser

from suanpan.arguments import Bool, String
from suanpan.components import Component as c
from suanpan.utils import image, npy


def showAsImage(data, temp="tmp", flag=None):
    imageType = "gif" if flag == "animated" else "png"
    filepath = os.path.join(temp, f"{time.time()}.{imageType}")
    image.save(filepath, data, flag=flag)
    showImage(filepath)


def showImage(filepath):
    url = "file://" + os.path.abspath(filepath)
    webbrowser.open(url)


@c.input(String(key="npy", required=True))
@c.param(String(key="toImage"))
@c.param(String(key="flag"))
@c.param(Bool(key="show", default=False))
def SPNpyTools(context):
    args = context.args

    data = npy.load(args.npy)
    if args.toImage:
        filepath = image.save(args.toImage, data, flag=args.flag)
        if args.show:
            showImage(filepath)
    else:
        import pdb  # pylint: disable=import-outside-toplevel

        pdb.set_trace()


if __name__ == "__main__":
    SPNpyTools()  # pylint: disable=no-value-for-parameter



# coding=utf-8
from __future__ import absolute_import, print_function

import copy
import functools

from suanpan import error
from suanpan.imports import imports
from suanpan.objects import HasName


class Proxy(HasName):
    TYPE_KEY = "type"
    DEFAULT_TYPE = None
    MAPPING = {}

    def __init__(self, *args, **kwargs):
        super(Proxy, self).__init__()
        self._backend = None
        self._args = args
        self._kwargs = kwargs
        setattr(self, self.TYPE_KEY, kwargs.get(self.TYPE_KEY, self.DEFAULT_TYPE))

    def __call__(self, *args, **kwargs):
        return self.backend(*args, **kwargs)  # pylint: disable=not-callable

    def __getattr__(self, key):
        return getattr(self.__getattribute__("backend"), key)

    def __copy__(self):
        return self.__class__(*self._args, **self._kwargs)

    def __deepcopy__(self, meno=None):
        return self.__class__(
            *copy.deepcopy(self._args, meno), **copy.deepcopy(self._kwargs, meno)
        )

    @property
    def backend(self):
        if self._backend is None:
            if getattr(self, self.TYPE_KEY, None) is None:
                raise error.ProxyError(f"{self.name} backend isn't set.")
            self.setBackend(*self._args, **self._kwargs)
        if isinstance(self._backend, functools.partial):
            self._backend = self._backend()
        return self._backend

    def setBackend(self, *args, **kwargs):
        backendType = kwargs.get(self.TYPE_KEY, getattr(self, self.TYPE_KEY, None))
        if not backendType:
            raise error.ProxyError(
                f"{self.name} backend type is required - {self.TYPE_KEY}: {backendType}"
            )
        BackendClass = self.MAPPING.get(backendType)
        if not BackendClass:
            raise error.ProxyError(
                f"{self.name} don't supported backend type {backendType}"
            )
        if isinstance(BackendClass, str):
            BackendClass = imports(BackendClass)
        _args = [*self._args, *args]
        _kwargs = {**self._kwargs, **kwargs}
        self._backend = self._setBackend(BackendClass, *_args, **_kwargs)
        setattr(self, self.TYPE_KEY, backendType)
        return self

    def setDefaultBackend(self, *args, **kwargs):
        kwargs.update({self.TYPE_KEY: self.DEFAULT_TYPE})
        return self.setBackend(*args, **kwargs)

    def _setBackend(self, BackendClass, *args, **kwargs):
        if self.TYPE_KEY in kwargs:
            kwargs.pop(self.TYPE_KEY)
        return functools.partial(BackendClass, *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

import functools
import json

import numpy as np

from suanpan import path


class SPEncoder(json.JSONEncoder):
    NUMPY_INT = (
        np.int_,
        np.intc,
        np.intp,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
    )
    NUMPY_FLOAT = (np.float_, np.float16, np.float32, np.float64)
    NUMPY_ARRAY = (np.ndarray,)

    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, self.NUMPY_INT):
            return int(obj)
        if isinstance(obj, self.NUMPY_FLOAT):
            return float(obj)
        if isinstance(obj, self.NUMPY_ARRAY):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


_load = functools.partial(json.load, encoding="utf-8")
_loads = functools.partial(json.loads, encoding="utf-8")
_dump = functools.partial(json.dump, ensure_ascii=False, cls=SPEncoder)
_dumps = functools.partial(json.dumps, ensure_ascii=False, cls=SPEncoder)


def _loadf(file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    with open(file, "r", encoding=encoding) as _file:
        return _load(_file, *args, **kwargs)


def _dumpf(s, file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    path.safeMkdirsForFile(file)
    with open(file, "w", encoding=encoding) as _file:
        return _dump(s, _file, *args, **kwargs)


def load(file, *args, **kwargs):
    _l = _loadf if isinstance(file, str) else _load
    return _l(file, *args, **kwargs)


def dump(s, file, *args, **kwargs):
    _d = _dumpf if isinstance(file, str) else _dump
    return _d(s, file, *args, **kwargs)


loads = _loads
dumps = _dumps
